/*****************************************************
DeleteIsFinal (safety property):
DeleteIsFinal checks that object cannot be Get after removal.
*****************************************************/

// event: initialize the monitor with the initial account balances for all clients when the system starts
event eSpec_DeleteIsFinal_Init;

/****************************************************
BankBalanceIsAlwaysCorrect checks the global invariant that the account balance communicated
to the client by the bank is always correct and there is no error on the banks side with the
implementation of the withdraw logic.
****************************************************/
spec DeleteIsFinal observes eShardReqDel, eShardReqPut, eShardResExists, eShardResDel, eShardResPut, eSpec_DeleteIsFinal_Init {
  var storage: set[int];
  var pendingPut: set[int];
  var pendingDel: set[int];

  start state Init {
    on eSpec_DeleteIsFinal_Init goto WaitForAnOp;
  }

  state WaitForAnOp {
    on eShardReqDel do (resp: tShardReqDel) {
        pendingDel += (resp.address);
    }
    on eShardReqPut do (resp: tShardReqPut) {
        pendingPut += (resp.obj.address);
    }
    on eShardResDel do (resp: tShardResDel) {
        assert resp.address in pendingDel;
        pendingDel -= resp.address;
        storage -= (resp.address);
    }
    on eShardResPut do (resp: tShardResPut) {
        assert resp.address in pendingPut;
        pendingPut -= resp.address;
        storage += (resp.address);
    }
    on eShardResExists do (resp: tShardResExists) {
        assert resp.found == resp.address in storage;
    }
  }
}