type tObject = (address: int, obj: int);
type tCacheReqPut = (source: machine, obj: tObject);
type tCacheReqDel = (source: machine, address: int);

type tCacheResPut = (address: int);
type tCacheResDel = (address: int);
event eCacheReqPut: tCacheReqPut;
event eCacheResPut: tCacheResPut;
event eCacheReqDel: tCacheReqDel;
event eCacheResDel: tCacheResDel;
event eFlushSingle;

machine Cache
{
    var objects: map[int, tObject];
    var b: Blobstor;
    var m: Metabase;

    start state Init {
        entry (input: (b: Blobstor, m: Metabase))
        {
            b = input.b;
            m = input.m;
            goto FlushLoop;
        }
    }
    state FlushLoop {
        on eFlushSingle do {
            var obj: tObject;
            var storageId: int;
            var address: int;
            if (sizeof(objects) > 0) {
                address = choose(objects);
                print format ("Start FLUSH, address={0}", address);
                obj = objects[address];

                send b, eBlobReqPut, (source = this, obj = obj);
                receive { case eBlobResPut: (r: tBlobResPut) { storageId = r.storageId; } }

                send m, eMetaReqUpdateId, (source = this, address = address, storageId = storageId);
                receive { case eMetaResUpdateId: (r: tMetaResUpdateId) {} }

                objects -= address;
                print format ("Finish FLUSH, address={0}, remaining={1}", address, sizeof(objects));
            }
        }

        on eCacheReqPut do (r: tCacheReqPut) {
            print format ("PUT address={0}", r.obj.address);
            if (!(r.obj.address in objects)) {
                objects += (r.obj.address, r.obj);
            }
            send r.source, eCacheResPut, (address = r.obj.address,);
            if ($) { send this, eFlushSingle; }
        }

        on eCacheReqDel do (r: tCacheReqDel) {
            print format ("DELETE address={0}", r.address);
            objects -= (r.address) ;
            send r.source, eCacheResDel, (address = r.address,);
            if ($) { send this, eFlushSingle; }
        }
    }
}

