type tShardReqPut = (source: Client, obj: tObject);
type tShardResPut = (address: int);
type tShardReqDel = (source: Client, address: int);
type tShardResDel = (address: int);
type tShardReqExists = (source: Client, address: int);
type tShardResExists = (address: int, found: bool);

event eShardReqPut: tShardReqPut;
event eShardResPut: tShardResPut;
event eShardReqDel: tShardReqDel;
event eShardResDel: tShardResDel;
event eShardReqExists: tShardReqExists;
event eShardResExists: tShardResExists;

machine Shard
{
    var c: Cache;
    var b: Blobstor;
    var m: Metabase;

    start state Init {
        entry (input: (b: Blobstor, m: Metabase))
        {
            b = input.b;
            m = input.m;
            c = new Cache((b = b, m = m));
            goto WorkLoop;
        }
    }
    state WorkLoop {
        on eShardReqPut do (r: tShardReqPut) {
            send c, eCacheReqPut, (source = this, obj = r.obj);
            receive { case eCacheResPut: (resp: tCacheResPut) {} }

            send m, eMetaReqUpdateId, (source = this, address = r.obj.address, storageId = 0);
            receive { case eMetaResUpdateId: (resp: tMetaResUpdateId) {}}

            send r.source, eShardResPut, (address = r.obj.address,);
        }

        on eShardReqDel do (r: tShardReqDel) {
//            send c, eCacheReqDel, (source = this, obj = r.obj);
//            receive { case eCacheResDel: (resp: tCacheResDel) {} }
            send b, eBlobReqDel, (source = this, address = r.address);
            receive { case eBlobResDel: (resp: tBlobResDel) {}}

            send m, eMetaReqDelete, (source = this, address = r.address);
            receive { case eMetaResDelete: (resp: tMetaResDelete) {}}

            send r.source, eShardResDel, (address = r.address,);
        }

        on eShardReqExists do (r: tShardReqExists) {
            var found: bool;
            send m, eMetaReqExists, (source = this, address = r.address);
            receive {
                case eMetaResExists: (resp: tMetaResExists) {
                    found = resp.found;
                }
            }
            send r.source, eShardResExists, (address = r.address, found = found);
        }
    }
}